@@ -1,6 +1,7 @@
 /* This file is part of the KDE libraries
    Copyright (C) 2005 Olivier Goffart <ogoffart at kde.org>
    Copyright (C) 2013-2015 Martin Klapetek <mklapetek@kde.org>
+   Copyright (C) 2017 Eike Hein <hein@kde.org>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -25,7 +26,10 @@
 #include <QDBusConnection>
 #include <QPointer>
 #include <QBuffer>
+#include <QFileInfo>
+#include <QDBusConnectionInterface>
 #include <KPluginLoader>
+#include <KPluginMetaData>
 
 #include "knotifyconfig.h"
 #include "knotificationplugin.h"
@@ -34,10 +38,13 @@
 #include "notifybylogfile.h"
 #include "notifybytaskbar.h"
 #include "notifybyexecute.h"
+#include "notifybyflatpak.h"
 #include "debug_p.h"
 
-#ifdef HAVE_PHONON4QT5
-#include "notifybyaudio.h"
+#if defined(HAVE_CANBERRA)
+#include "notifybyaudio_canberra.h"
+#elif defined(HAVE_PHONON4QT5)
+#include "notifybyaudio_phonon.h"
 #endif
 
 #ifdef HAVE_SPEECH
@@ -46,13 +53,15 @@
 
 typedef QHash<QString, QString> Dict;
 
-struct KNotificationManager::Private {
+struct Q_DECL_HIDDEN KNotificationManager::Private {
     QHash<int, KNotification *> notifications;
     QHash<QString, KNotificationPlugin *> notifyPlugins;
 
     // incremental ids for notifications
     int notifyIdCounter;
     QStringList dirtyConfigCache;
+    bool inSandbox = false;
+    bool portalDBusServiceExists = false;
 };
 
 class KNotificationManagerSingleton
@@ -74,42 +83,25 @@ KNotificationManager::KNotificationManager()
     d->notifyIdCounter = 0;
     qDeleteAll(d->notifyPlugins);
     d->notifyPlugins.clear();
-    addPlugin(new NotifyByPopup(this));
-    addPlugin(new NotifyByExecute(this));
-    //FIXME: port and reenable
-//     addPlugin(new NotifyByLogfile(this));
 
-#ifdef HAVE_PHONON4QT5
-    addPlugin(new NotifyByAudio(this));
-#endif
-
-    addPlugin(new NotifyByTaskbar(this));
-
-#ifdef HAVE_SPEECH
-    addPlugin(new NotifyByTTS(this));
-#endif
-
-    QList<QObject*> plugins = KPluginLoader::instantiatePlugins(QStringLiteral("knotification/notifyplugins"),
-                                                                std::function<bool(const KPluginMetaData &)>(),
-                                                                this);
-
-    Q_FOREACH (QObject *plugin, plugins) {
-        KNotificationPlugin *notifyPlugin = qobject_cast<KNotificationPlugin*>(plugin);
-        if (notifyPlugin) {
-            addPlugin(notifyPlugin);
-        } else {
-            // not our/valid plugin, so delete the created object
-            plugin->deleteLater();
+    if (!qEnvironmentVariableIsEmpty("XDG_RUNTIME_DIR")) {
+        const QByteArray runtimeDir = qgetenv("XDG_RUNTIME_DIR");
+        if (!runtimeDir.isEmpty()) {
+            d->inSandbox = QFileInfo::exists(QFile::decodeName(runtimeDir) + QLatin1String("/flatpak-info"));
         }
     }
 
+    if (d->inSandbox) {
+        QDBusConnectionInterface *interface = QDBusConnection::sessionBus().interface();
+        d->portalDBusServiceExists = interface->isServiceRegistered(QStringLiteral("org.freedesktop.portal.Desktop"));
+    }
+
     QDBusConnection::sessionBus().connect(QString(),
                                           QStringLiteral("/Config"),
                                           QStringLiteral("org.kde.knotification"),
                                           QStringLiteral("reparseConfiguration"),
                                           this,
                                           SLOT(reparseConfiguration(QString)));
-
 }
 
 KNotificationManager::~KNotificationManager()
@@ -117,11 +109,109 @@ KNotificationManager::~KNotificationManager()
     delete d;
 }
 
-void KNotificationManager::addPlugin(KNotificationPlugin *notifyPlugin)
+KNotificationPlugin *KNotificationManager::pluginForAction(const QString &action)
 {
-    d->notifyPlugins[notifyPlugin->optionName()] = notifyPlugin;
-    connect(notifyPlugin, SIGNAL(finished(KNotification*)), this, SLOT(notifyPluginFinished(KNotification*)));
-    connect(notifyPlugin, SIGNAL(actionInvoked(int, int)), this, SLOT(notificationActivated(int, int)));
+    KNotificationPlugin *plugin = d->notifyPlugins.value(action);
+
+    // We already loaded a plugin for this action.
+    if (plugin) {
+        return plugin;
+    }
+
+    auto addPlugin = [this](KNotificationPlugin *plugin) {
+        d->notifyPlugins[plugin->optionName()] = plugin;
+        connect(plugin, &KNotificationPlugin::finished,
+                this, &KNotificationManager::notifyPluginFinished);
+        connect(plugin, &KNotificationPlugin::actionInvoked,
+                this, &KNotificationManager::notificationActivated);
+    };
+
+    // Load plugin.
+    // We have a series of built-ins up first, and fall back to trying
+    // to instantiate an externally supplied plugin.
+    if (action == QLatin1String("Popup")) {
+            if (d->inSandbox && d->portalDBusServiceExists) {
+                plugin = new NotifyByFlatpak(this);
+            } else {
+                plugin = new NotifyByPopup(this);
+            }
+
+        addPlugin(plugin);
+    } else if (action == QLatin1String("Taskbar")) {
+        plugin = new NotifyByTaskbar(this);
+        addPlugin(plugin);
+    } else if (action == QLatin1String("Sound")) {
+#if defined(HAVE_PHONON4QT5) || defined(HAVE_CANBERRA)
+        plugin = new NotifyByAudio(this);
+        addPlugin(plugin);
+#endif
+    } else if (action == QLatin1String("Execute")) {
+        plugin = new NotifyByExecute(this);
+        addPlugin(plugin);
+    } else if (action == QLatin1String("Logfile")) {
+        plugin = new NotifyByLogfile(this);
+        addPlugin(plugin);
+    } else if (action == QLatin1String("TTS")) {
+#ifdef HAVE_SPEECH
+        plugin = new NotifyByTTS(this);
+        addPlugin(plugin);
+#endif
+    } else {
+        bool pluginFound = false;
+
+        QList<QObject*> plugins = KPluginLoader::instantiatePlugins(QStringLiteral("knotification/notifyplugins"),
+            [&action, &pluginFound](const KPluginMetaData &data) {
+                // KPluginLoader::instantiatePlugins loops over the plugins it
+                // found and calls this function to determine whether to
+                // instantiate them. We use a `pluginFound` var outside the
+                // lambda to break out of the loop once we got a match.
+                // The reason we can't just use KPluginLoader::findPlugins,
+                // loop over the meta data and instantiate only one plugin
+                // is because the X-KDE-KNotification-OptionName field is
+                // optional (see TODO note below) and the matching plugin
+                // may be among the plugins which don't have it.
+                if (pluginFound) {
+                    return false;
+                }
+
+                const QJsonObject &rawData = data.rawData();
+
+                // This field is new-ish and optional. If it's not set we always
+                // instantiate the plugin, unless we already got a match.
+                // TODO KF6: Require X-KDE-KNotification-OptionName be set and
+                // reject plugins without it.
+                if (rawData.contains(QStringLiteral("X-KDE-KNotification-OptionName"))) {
+                    if (rawData.value(QStringLiteral("X-KDE-KNotification-OptionName")) == action) {
+                        pluginFound = true;
+                    } else {
+                        return false;
+                    }
+                }
+
+                return true;
+            },
+            this);
+
+        Q_FOREACH (QObject *pluginObj, plugins) {
+            KNotificationPlugin *notifyPlugin = qobject_cast<KNotificationPlugin*>(pluginObj);
+
+            if (notifyPlugin) {
+                // We try to avoid unnecessary instantiations (see above), but
+                // when they happen keep the resulting plugins around.
+                addPlugin(notifyPlugin);
+
+                // Get ready to return the plugin we got asked for.
+                if (notifyPlugin->optionName() == action) {
+                    plugin = notifyPlugin;
+                }
+            } else {
+                // Not our/valid plugin, so delete the created object.
+                pluginObj->deleteLater();
+            }
+        }
+    }
+
+    return plugin;
 }
 
 void KNotificationManager::notifyPluginFinished(KNotification *notification)
@@ -175,7 +265,7 @@ void KNotificationManager::close(int id, bool force)
         KNotifyConfig notifyConfig(n->appName(), n->contexts(), n->eventId());
         QString notifyActions = notifyConfig.readEntry(QStringLiteral("Action"));
 
-        Q_FOREACH (const QString &action, notifyActions.split('|')) {
+        Q_FOREACH (const QString &action, notifyActions.split(QLatin1Char('|'))) {
             if (!d->notifyPlugins.contains(action)) {
                 qCDebug(LOG_KNOTIFICATIONS) << "No plugin for action" << action;
                 continue;
@@ -195,23 +285,25 @@ int KNotificationManager::notify(KNotification *n)
         d->dirtyConfigCache.removeOne(n->appName());
     }
 
-    QString notifyActions = notifyConfig.readEntry(QStringLiteral("Action"));
+    const QString notifyActions = notifyConfig.readEntry(QStringLiteral("Action"));
 
     if (notifyActions.isEmpty() || notifyActions == QLatin1String("None")) {
         // this will cause KNotification closing itself fast
+        n->ref();
         n->deref();
         return -1;
     }
 
     d->notifications.insert(d->notifyIdCounter, n);
 
-    Q_FOREACH (const QString &action, notifyActions.split('|')) {
-        if (!d->notifyPlugins.contains(action)) {
+    Q_FOREACH (const QString &action, notifyActions.split(QLatin1Char('|'))) {
+        KNotificationPlugin *notifyPlugin = pluginForAction(action);
+
+        if (!notifyPlugin) {
             qCDebug(LOG_KNOTIFICATIONS) << "No plugin for action" << action;
             continue;
         }
 
-        KNotificationPlugin *notifyPlugin = d->notifyPlugins[action];
         n->ref();
         qCDebug(LOG_KNOTIFICATIONS) << "Calling notify on" << notifyPlugin->optionName();
         notifyPlugin->notify(n, &notifyConfig);
@@ -223,13 +315,6 @@ int KNotificationManager::notify(KNotification *n)
 
 void KNotificationManager::update(KNotification *n)
 {
-    QByteArray pixmapData;
-    if (!n->pixmap().isNull()) {
-        QBuffer buffer(&pixmapData);
-        buffer.open(QIODevice::WriteOnly);
-        n->pixmap().save(&buffer, "PNG");
-    }
-
     KNotifyConfig notifyConfig(n->appName(), n->contexts(), n->eventId());
 
     Q_FOREACH (KNotificationPlugin *p, d->notifyPlugins) {
@@ -239,15 +324,6 @@ void KNotificationManager::update(KNotification *n)
 
 void KNotificationManager::reemit(KNotification *n)
 {
-    QVariantList contextList;
-    typedef QPair<QString, QString> Context;
-    foreach (const Context &ctx, n->contexts()) {
-        QVariantList vl;
-        vl << ctx.first << ctx.second;
-        contextList << vl;
-    }
-
-
     notify(n);
 }
 
